"""Contains opsin models, parameters, and OptogeneticIntervention device"""
from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Tuple, Any

from brian2 import Synapses, NeuronGroup
from brian2.units import (
    mm,
    mm2,
    nmeter,
    meter,
    kgram,
    Quantity,
    second,
    ms,
    second,
    psiemens,
    mV,
    volt,
    amp,
    mwatt,
)
from brian2.units.allunits import meter2, radian
import brian2.units.unitsafefunctions as usf
from brian2.core.base import BrianObjectException
import numpy as np
import matplotlib
from matplotlib import colors
from matplotlib.artist import Artist
from matplotlib.collections import PathCollection

from cleosim.utilities import wavelength_to_rgb
from cleosim.stimulators import Stimulator


ChR2_four_state = {
    "g0": 114000 * psiemens,
    "gamma": 0.00742,
    "phim": 2.33e17 / mm2 / second,  # *photon, not in Brian2
    "k1": 4.15 / ms,
    "k2": 0.868 / ms,
    "p": 0.833,
    "Gf0": 0.0373 / ms,
    "kf": 0.0581 / ms,
    "Gb0": 0.0161 / ms,
    "kb": 0.063 / ms,
    "q": 1.94,
    "Gd1": 0.105 / ms,
    "Gd2": 0.0138 / ms,
    "Gr0": 0.00033 / ms,
    "E": 0 * mV,
    "v0": 43 * mV,
    "v1": 17.1 * mV,
}
"""Parameters for the 4-state ChR2 model.

Taken from try.projectpyrho.org's default 4-state params.
"""


class OpsinModel(ABC):
    """Base class for opsin model"""

    model: str
    """Basic Brian model equations string.
    
    Should contain a `rho_rel` term reflecting relative expression 
    levels. Will likely also contain special NeuronGroup-dependent
    symbols such as V_VAR_NAME to be replaced on injection in 
    :meth:`get_modified_model_for_ng`."""

    params: dict
    """Parameter values for model, passed in as a namespace dict"""

    @abstractmethod
    def get_modified_model_for_ng(
        self, neuron_group: NeuronGroup, injct_params: dict
    ) -> str:
        """Adapt model for given neuron group on injection

        This is mainly to enable the specification of variable names
        differently for each neuron group.

        Parameters
        ----------
        neuron_group : NeuronGroup
            NeuronGroup this opsin model is being connected to
        injct_params : dict
            kwargs passed in on injection, could contain variable
            names to plug into the model

        Returns
        -------
        str
            A string containing modified model equations for use
            in :attr:`~cleosim.opto.OptogeneticIntervention.opto_syns`
        """
        pass

    def init_opto_syn_vars(self, opto_syn: Synapses) -> None:
        """Initializes appropriate variables in Synapses implementing the model

        Can also be used to reset the variables.

        Parameters
        ----------
        opto_syn : Synapses
            The synapses object implementing this model
        """
        pass


class MarkovModel(OpsinModel):
    """Base class for Markov state models à la Evans et al., 2016"""

    def __init__(self, params: dict) -> None:
        """
        Parameters
        ----------
        params : dict
            dict defining params in the :attr:`model`
        """
        super().__init__()
        self.params = params

    def _check_vars_on_injection(self, neuron_group, injct_params):
        Iopto_var_name = injct_params.get("Iopto_var_name", "Iopto")
        v_var_name = injct_params.get("v_var_name", "v")
        for variable, unit in zip([v_var_name, Iopto_var_name], [volt, amp]):
            if (
                variable not in neuron_group.variables
                or neuron_group.variables[variable].unit != unit
            ):
                raise BrianObjectException(
                    (
                        f"{variable} : {unit.name} needed in the model of NeuronGroup"
                        f"{neuron_group.name} to connect OptogeneticIntervention."
                    ),
                    neuron_group,
                )

    def get_modified_model_for_ng(
        self, neuron_group: NeuronGroup, injct_params: dict
    ) -> str:
        self._check_vars_on_injection(neuron_group, injct_params)
        Iopto_var_name = injct_params.get("Iopto_var_name", "Iopto")
        v_var_name = injct_params.get("v_var_name", "v")
        # opsin synapse model needs modified names
        return self.model.replace("IOPTO_VAR_NAME", Iopto_var_name).replace(
            "V_VAR_NAME", v_var_name
        )


class FourStateModel(MarkovModel):
    """4-state model from PyRhO (Evans et al. 2016).

    rho_rel is channel density relative to standard model fit;
    modifying it post-injection allows for heterogeneous opsin expression.

    IOPTO_VAR_NAME and V_VAR_NAME are substituted on injection.
    """

    model: str = """
        dC1/dt = Gd1*O1 + Gr0*C2 - Ga1*C1 : 1 (clock-driven)
        dO1/dt = Ga1*C1 + Gb*O2 - (Gd1+Gf)*O1 : 1 (clock-driven)
        dO2/dt = Ga2*C2 + Gf*O1 - (Gd2+Gb)*O2 : 1 (clock-driven)
        C2 = 1 - C1 - O1 - O2 : 1
        # dC2/dt = Gd2*O2 - (Gr0+Ga2)*C2 : 1 (clock-driven)

        Theta = int(phi > 0*phi) : 1
        Hp = Theta * phi**p/(phi**p + phim**p) : 1
        Ga1 = k1*Hp : hertz
        Ga2 = k2*Hp : hertz
        Hq = Theta * phi**q/(phi**q + phim**q) : 1
        Gf = kf*Hq + Gf0 : hertz
        Gb = kb*Hq + Gb0 : hertz

        fphi = O1 + gamma*O2 : 1
        fv = (1 - exp(-(V_VAR_NAME_post-E)/v0)) / -2 : 1

        IOPTO_VAR_NAME_post = -g0*fphi*fv*(V_VAR_NAME_post-E)*rho_rel : ampere (summed)
        rho_rel : 1
    """

    def init_opto_syn_vars(self, opto_syn: Synapses) -> None:
        for varname, value in {"Irr0": 0, "C1": 1, "O1": 0, "O2": 0}.items():
            setattr(opto_syn, varname, value)


class ProportionalCurrentModel(OpsinModel):
    """A simple model delivering current proportional to light intensity"""

    model: str = """
        IOPTO_VAR_NAME_post = gain * Irr * rho_rel : IOPTO_UNIT (summed)
        rho_rel : 1
    """

    def __init__(self, Iopto_per_mW_per_mm2: Quantity) -> None:
        """
        Parameters
        ----------
        Iopto_per_mW_per_mm2 : Quantity
            How much current (in amps or unitless, depending on neuron model) to
            deliver per mW/mm2
        """
        self.params = {"gain": Iopto_per_mW_per_mm2 / (mwatt / mm2)}
        if isinstance(Iopto_per_mW_per_mm2, Quantity):
            self._Iopto_dim = Iopto_per_mW_per_mm2.get_best_unit().dim
            self._Iopto_dim_name = self._Iopto_dim._str_representation(python_code=True)
        else:
            self._Iopto_dim = radian.dim
            self._Iopto_dim_name = self._Iopto_dim._str_representation(
                python_code=False
            )

    def get_modified_model_for_ng(
        self, neuron_group: NeuronGroup, injct_params: dict
    ) -> str:
        Iopto_var_name = injct_params.get("Iopto_var_name", "Iopto")
        if (
            Iopto_var_name not in neuron_group.variables
            or neuron_group.variables[Iopto_var_name].dim != self._Iopto_dim
        ):
            raise BrianObjectException(
                (
                    f"{Iopto_var_name} : {self._Iopto_dim_name} needed in the model of NeuronGroup"
                    f"{neuron_group.name} to connect OptogeneticIntervention."
                ),
                neuron_group,
            )
        return self.model.replace("IOPTO_VAR_NAME", Iopto_var_name).replace(
            "IOPTO_UNIT", self._Iopto_dim_name
        )


default_blue = {
    "R0": 0.1 * mm,  # optical fiber radius
    "NAfib": 0.37,  # optical fiber numerical aperture
    "wavelength": 473 * nmeter,
    # NOTE: the following depend on wavelength and tissue properties and thus would be different for another wavelength
    "K": 0.125 / mm,  # absorbance coefficient
    "S": 7.37 / mm,  # scattering coefficient
    "ntis": 1.36,  # tissue index of refraction
}
"""Light parameters for 473 nm wavelength delivered via an optic fiber.

From Foutz et al., 2012"""


class OptogeneticIntervention(Stimulator):
    """Enables optogenetic stimulation of the network.

    Essentially "transfects" neurons and provides a light source.
    Under the hood, it delivers current via a Brian :class:`~brian2.synapses.synapses.Synapses`
    object.

    Requires neurons to have 3D spatial coordinates already assigned.
    Also requires that the neuron model has a current term
    (by default Iopto) which is assumed to be positive (unlike the
    convention in many opsin modeling papers, where the current is
    described as negative).

    See :meth:`connect_to_neuron_group` for optional keyword parameters
    that can be specified when calling
    :meth:`cleosim.CLSimulator.inject_stimulator`.
    """

    opto_syns: dict[str, Synapses]
    """Stores the synapse objects implementing the opsin model,
    with NeuronGroup name keys and Synapse values."""

    max_Irr0_mW_per_mm2: float
    """The maximum irradiance the light source can emit.
    
    Usually determined by hardware in a real experiment."""

    max_Irr0_mW_per_mm2_viz: float
    """Maximum irradiance for visualization purposes. 
    
    i.e., the level at or above which the light appears maximally bright.
    Only relevant in video visualization.
    """

    def __init__(
        self,
        name: str,
        opsin_model: OpsinModel,
        light_model_params: dict,
        location: Quantity = (0, 0, 0) * mm,
        direction: Tuple[float, float, float] = (0, 0, 1),
        max_Irr0_mW_per_mm2: float = None,
        save_history: bool = False,
    ):
        """
        Parameters
        ----------
        name : str
            Unique identifier for stimulator
        opsin_model : OpsinModel
            OpsinModel object defining how light affects target
            neurons. See :class:`FourStateModel` and :class:`ProportionalCurrentModel`
            for examples.
        light_model_params : dict
            Parameters for the light propagation model in Foutz et al., 2012.
            See :attr:`default_blue` for an example.
        location : Quantity, optional
            (x, y, z) coords with Brian unit specifying where to place
            the base of the light source, by default (0, 0, 0)*mm
        direction : Tuple[float, float, float], optional
            (x, y, z) vector specifying direction in which light
            source is pointing, by default (0, 0, 1)
        max_Irr0_mW_per_mm2 : float, optional
            Set :attr:`max_Irr0_mW_per_mm2`.
        save_history : bool, optional
            Determines whether :attr:`~values` and :attr:`~t_ms` are saved.
        """
        super().__init__(name, 0, save_history)
        self.opsin_model = opsin_model
        self.light_model_params = light_model_params
        self.location = location
        # direction unit vector
        self.dir_uvec = (direction / np.linalg.norm(direction)).reshape((3, 1))
        self.opto_syns = {}
        self.max_Irr0_mW_per_mm2 = max_Irr0_mW_per_mm2
        self.max_Irr0_mW_per_mm2_viz = None

    def _Foutz12_transmittance(self, r, z, scatter=True, spread=True, gaussian=True):
        """Foutz et al. 2012 transmittance model: Gaussian cone with Kubelka-Munk propagation"""

        if spread:
            # divergence half-angle of cone
            theta_div = np.arcsin(
                self.light_model_params["NAfib"] / self.light_model_params["ntis"]
            )
            Rz = self.light_model_params["R0"] + z * np.tan(
                theta_div
            )  # radius as light spreads("apparent radius" from original code)
            C = (self.light_model_params["R0"] / Rz) ** 2
        else:
            Rz = self.light_model_params["R0"]  # "apparent radius"
            C = 1

        if gaussian:
            G = 1 / np.sqrt(2 * np.pi) * np.exp(-2 * (r / Rz) ** 2)
        else:
            G = 1

        def kubelka_munk(dist):
            S = self.light_model_params["S"]
            a = 1 + self.light_model_params["K"] / S
            b = np.sqrt(a**2 - 1)
            dist = np.sqrt(r**2 + z**2)
            return b / (a * np.sinh(b * S * dist) + b * np.cosh(b * S * dist))

        M = kubelka_munk(np.sqrt(r**2 + z**2)) if scatter else 1

        T = G * C * M
        return T

    def _get_rz_for_xyz(self, x, y, z):
        """Assumes x, y, z already have units"""

        def flatten_if_needed(var):
            if len(var.shape) != 1:
                return var.flatten()
            else:
                return var

        # have to add unit back on since it's stripped by vstack
        coords = (
            np.vstack(
                [flatten_if_needed(x), flatten_if_needed(y), flatten_if_needed(z)]
            ).T
            * meter
        )
        rel_coords = coords - self.location  # relative to fiber location
        # must use brian2's dot function for matrix multiply to preserve
        # units correctly.
        zc = usf.dot(rel_coords, self.dir_uvec)  # distance along cylinder axis
        # just need length (norm) of radius vectors
        # not using np.linalg.norm because it strips units
        r = np.sqrt(np.sum((rel_coords - usf.dot(zc, self.dir_uvec.T)) ** 2, axis=1))
        r = r.reshape((-1, 1))
        return r, zc

    def connect_to_neuron_group(
        self, neuron_group: NeuronGroup, **kwparams: Any
    ) -> None:
        """Configure opsin and light source to stimulate given neuron group.

        Parameters
        ----------
        neuron_group : NeuronGroup
            The neuron group to stimulate with the given opsin and light source

        Keyword args
        ------------
        p_expression : float
            Probability (0 <= p <= 1) that a given neuron in the group
            will express the opsin. 1 by default.
        rho_rel : float
            The expression level, relative to the standard model fit,
            of the opsin. 1 by default. For heterogeneous expression,
            this would have to be modified in the opsin synapse post-injection,
            e.g., ``opto.opto_syns["neuron_group_name"].rho_rel = ...``
        Iopto_var_name : str
            The name of the variable in the neuron group model representing
            current from the opsin
        v_var_name : str
            The name of the variable in the neuron group model representing
            membrane potential
        """
        # get modified opsin model string (i.e., with names/units specified)
        modified_opsin_model = self.opsin_model.get_modified_model_for_ng(
            neuron_group, kwparams
        )

        # fmt: off
        # Ephoton = h*c/lambda
        E_photon = (
            6.63e-34 * meter2 * kgram / second
            * 2.998e8 * meter / second
            / self.light_model_params["wavelength"]
        )
        # fmt: on

        light_model = """
            Irr = Irr0*T : watt/meter**2
            Irr0 : watt/meter**2 
            T : 1
            phi = Irr / Ephoton : 1/second/meter**2
        """

        opto_syn = Synapses(
            neuron_group,
            model=modified_opsin_model + light_model,
            namespace=self.opsin_model.params,
            name=f"synapses_{self.name}_{neuron_group.name}",
            method="rk2",
        )
        opto_syn.namespace["Ephoton"] = E_photon

        p_expression = kwparams.get("p_expression", 1)
        if p_expression == 1:
            opto_syn.connect(j="i")
        else:
            opto_syn.connect(condition="i==j", p=p_expression)

        self.opsin_model.init_opto_syn_vars(opto_syn)

        # relative channel density
        opto_syn.rho_rel = kwparams.get("rho_rel", 1)
        # calculate transmittance coefficient for each point
        r, z = self._get_rz_for_xyz(neuron_group.x, neuron_group.y, neuron_group.z)
        T = self._Foutz12_transmittance(r, z).flatten()
        # reduce to subset expressing opsin before assigning
        T = [T[k] for k in opto_syn.i]

        opto_syn.T = T

        self.opto_syns[neuron_group.name] = opto_syn
        self.brian_objects.add(opto_syn)

    def add_self_to_plot(self, ax, axis_scale_unit) -> PathCollection:
        # show light with point field, assigning r and z coordinates
        # to all points
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()
        zlim = ax.get_zlim()
        x = np.linspace(xlim[0], xlim[1], 50)
        y = np.linspace(ylim[0], ylim[1], 50)
        z = np.linspace(zlim[0], zlim[1], 50)
        x, y, z = np.meshgrid(x, y, z) * axis_scale_unit

        r, zc = self._get_rz_for_xyz(x, y, z)
        T = self._Foutz12_transmittance(r, zc)
        # filter out points with <0.001 transmittance to make plotting faster
        plot_threshold = 0.001
        idx_to_plot = T[:, 0] >= plot_threshold
        x = x.flatten()[idx_to_plot]
        y = y.flatten()[idx_to_plot]
        z = z.flatten()[idx_to_plot]
        T = T[idx_to_plot, 0]
        point_cloud = ax.scatter(
            x / axis_scale_unit,
            y / axis_scale_unit,
            z / axis_scale_unit,
            c=T,
            cmap=self._alpha_cmap_for_wavelength(),
            marker=",",
            edgecolors="none",
            label=self.name,
        )
        handles = ax.get_legend().legendHandles
        c = wavelength_to_rgb(self.light_model_params["wavelength"] / nmeter)
        opto_patch = matplotlib.patches.Patch(color=c, label=self.name)
        handles.append(opto_patch)
        ax.legend(handles=handles)
        return [point_cloud]

    def update_artists(
        self, artists: list[Artist], value, *args, **kwargs
    ) -> list[Artist]:
        self._prev_value = getattr(self, "_prev_value", None)
        if value == self._prev_value:
            return []

        assert len(artists) == 1
        point_cloud = artists[0]

        if self.max_Irr0_mW_per_mm2_viz is not None:
            max_Irr0 = self.max_Irr0_mW_per_mm2_viz
        elif self.max_Irr0_mW_per_mm2 is not None:
            max_Irr0 = self.max_Irr0_mW_per_mm2
        else:
            raise Exception(
                f"OptogeneticIntervention '{self.name}' needs max_Irr0_mW_per_mm2_viz "
                "or max_Irr0_mW_per_mm2 "
                "set to visualize light intensity."
            )

        intensity = value / max_Irr0 if value <= max_Irr0 else max_Irr0
        point_cloud.set_cmap(self._alpha_cmap_for_wavelength(intensity))
        return [point_cloud]

    def update(self, Irr0_mW_per_mm2: float):
        """Set the light intensity, in mW/mm2 (without unit)

        Parameters
        ----------
        Irr0_mW_per_mm2 : float
            Desired light intensity for light source

        Raises
        ------
        ValueError
            When intensity is negative
        """
        if Irr0_mW_per_mm2 < 0:
            raise ValueError(f"{self.name}: light intensity Irr0 must be nonnegative")
        if (
            self.max_Irr0_mW_per_mm2 is not None
            and Irr0_mW_per_mm2 > self.max_Irr0_mW_per_mm2
        ):
            Irr0_mW_per_mm2 = self.max_Irr0_mW_per_mm2
        super().update(Irr0_mW_per_mm2)
        for opto_syn in self.opto_syns.values():
            opto_syn.Irr0 = Irr0_mW_per_mm2 * mwatt / mm2

    def reset(self, **kwargs):
        for opto_syn in self.opto_syns.values():
            self.opsin_model.init_opto_syn_vars(opto_syn)

    def _alpha_cmap_for_wavelength(self, intensity=0.5):
        c = wavelength_to_rgb(self.light_model_params["wavelength"] / nmeter)
        c_clear = (*c, 0)
        c_opaque = (*c, 0.6 * intensity)
        return colors.LinearSegmentedColormap.from_list(
            "incr_alpha", [(0, c_clear), (1, c_opaque)]
        )



"""Contains functions for assigning neuron coordinates and visualizing"""

from __future__ import annotations
from typing import Tuple

from brian2 import mm, meter, Unit
from brian2.groups.group import Group
from brian2.groups.neurongroup import NeuronGroup
from brian2.units.fundamentalunits import get_dimensions
import numpy as np

from cleosim.utilities import get_orth_vectors_for_v, modify_model_with_eqs


def assign_coords_grid_rect_prism(
    neuron_group: NeuronGroup,
    xlim: Tuple[float, float],
    ylim: Tuple[float, float],
    zlim: Tuple[float, float],
    shape: Tuple[int, int, int],
    unit: Unit = mm,
) -> None:
    """Assign grid coordinates to neurons in a rectangular grid

    Parameters
    ----------
    neuron_group : NeuronGroup
        The neuron group to assign coordinates to
    xlim : Tuple[float, float]
        xmin, xmax, with no unit
    ylim : Tuple[float, float]
        ymin, ymax, with no unit
    zlim : Tuple[float, float]
        zmin, zmax with no unit
    shape : Tuple[int, int, int]
        n_x, n_y, n_z tuple representing the shape of the resulting grid
    unit : Unit, optional
        Brian unit determining what scale to use for coordinates, by default mm

    Raises
    ------
    ValueError
        When the shape is incompatible with the number of neurons in the group
    """
    num_grid_elements = np.product(shape)
    if num_grid_elements != len(neuron_group):
        raise ValueError(
            f"Number of elements specified in shape ({num_grid_elements}) "
            f"does not match the number of neurons in the group ({len(neuron_group)})."
        )

    x = np.linspace(xlim[0], xlim[1], shape[0])
    y = np.linspace(ylim[0], ylim[1], shape[1])
    z = np.linspace(zlim[0], zlim[1], shape[2])

    x, y, z = np.meshgrid(x, y, z)
    assign_coords(neuron_group, x, y, z)


def assign_coords_rand_rect_prism(
    neuron_group: NeuronGroup,
    xlim: Tuple[float, float],
    ylim: Tuple[float, float],
    zlim: Tuple[float, float],
    unit: Unit = mm,
) -> None:
    """Assign random coordinates to neurons within a rectangular prism

    Parameters
    ----------
    neuron_group : NeuronGroup
        neurons to assign coordinates to
    xlim : Tuple[float, float]
        xmin, xmax without unit
    ylim : Tuple[float, float]
        ymin, ymax without unit
    zlim : Tuple[float, float]
        zmin, zmax without unit
    unit : Unit, optional
        Brian unit to specify scale implied in limits, by default mm
    """
    x = (xlim[1] - xlim[0]) * np.random.random(len(neuron_group)) + xlim[0]
    y = (ylim[1] - ylim[0]) * np.random.random(len(neuron_group)) + ylim[0]
    z = (zlim[1] - zlim[0]) * np.random.random(len(neuron_group)) + zlim[0]
    assign_coords(neuron_group, x, y, z)


def assign_coords_rand_cylinder(
    neuron_group: NeuronGroup,
    xyz_start: Tuple[float, float, float],
    xyz_end: Tuple[float, float, float],
    radius: float,
    unit: Unit = mm,
) -> None:
    """Assign random coordinates within a cylinder.

    *DON'T USE YET*: spacing isn't uniform, clustering around
    the center since a random height and radius is chosen for
    each neuron

    Parameters
    ----------
    neuron_group : NeuronGroup
        neurons to assign coordinates to
    xyz_start : Tuple[float, float, float]
        starting position of cylinder without unit
    xyz_end : Tuple[float, float, float]
        ending position of cylinder without unit
    radius : float
        radius of cylinder without unit
    unit : Unit, optional
        Brian unit to scale other params, by default mm
    """
    xyz_start = np.array(xyz_start)
    xyz_end = np.array(xyz_end)
    rs = radius * np.random.random(len(neuron_group))
    thetas = 2 * np.pi * np.random.random(len(neuron_group))
    cyl_length = np.linalg.norm(xyz_end - xyz_start)
    z_cyls = cyl_length * np.random.random(len(neuron_group))

    c = np.reshape(
        (xyz_end - xyz_start) / cyl_length, (-1, 1)
    )  # unit vector in direction of cylinder

    r1, r2 = get_orth_vectors_for_v(c)

    def r_unit_vecs(thetas):
        r1s = np.repeat(r1, len(thetas), 0)
        r2s = np.repeat(r2, len(thetas), 0)
        cosines = np.reshape(np.cos(thetas), (len(thetas), 1))
        sines = np.reshape(np.sin(thetas), (len(thetas), 1))
        return r1s * cosines + r2s * sines

    rs = np.reshape(rs, (len(rs), 1))
    points = np.reshape(xyz_start, (1, 3)) + (c * z_cyls).T + rs * r_unit_vecs(thetas)

    assign_coords(neuron_group, points[:, 0], points[:, 1], points[:, 2], unit)


def assign_coords(neuron_group: NeuronGroup, x: np.ndarray, y: np.ndarray, z: np.ndarray, unit: Unit = mm):
    """Assign arbitrary coordinates to neuron group.

    Parameters
    ----------
    neuron_group : NeuronGroup
        neurons to be assigned coordinates
    x : np.ndarray
        x positions to assign (preferably 1D with no unit)
    y : np.ndarray
        y positions to assign (preferably 1D with no unit)
    z : np.ndarray
        z positions to assign (preferably 1D with no unit)
    unit : Unit, optional
        Brian unit determining what scale to use for coordinates, by default mm
    """
    _init_variables(neuron_group)
    neuron_group.x = np.reshape(x, (-1,)) * unit
    neuron_group.y = np.reshape(y, (-1,)) * unit
    neuron_group.z = np.reshape(z, (-1,)) * unit


def _init_variables(group: Group):
    for dim_name in ["x", "y", "z"]:
        if hasattr(group, dim_name):
            setattr(group, dim_name, 0 * mm)
        else:
            if type(group) == NeuronGroup:
                modify_model_with_eqs(group, f"{dim_name}: meter")
            elif issubclass(type(group), Group):
                group.variables.add_array(
                    dim_name,
                    size=group._N,
                    dimensions=get_dimensions(meter),
                    dtype=float,
                    constant=True,
                    scalar=False,
                )
            else:
                raise NotImplementedError(
                    "Coordinate assignment only implemented for brian2.Group objects"
                )



from typing import Any
import numpy as np
from nptyping import NDArray
from cleosim.processing.base import ProcessingBlock


class FiringRateEstimator(ProcessingBlock):
    """Exponential filter to estimate firing rate.

    Requires `sample_time_ms` kwarg when process is called.
    """

    def __init__(self, tau_ms: float, sample_period_ms: float, **kwargs):
        """
        Parameters
        ----------
        tau_ms : float
            Time constant of filter
        sample_period_ms : float
            Sampling period in milliseconds
        """
        super().__init__(**kwargs)
        self.tau_s = tau_ms / 1000
        self.T_s = sample_period_ms / 1000
        self.alpha = np.exp(-sample_period_ms / tau_ms)
        self.prev_rate = None
        self.prev_time_ms = None

    def compute_output(
        self, input: NDArray[(Any,), np.uint], **kwargs
    ) -> NDArray[(Any,), float]:
        """Estimate firing rate given past and current spikes.

        Parameters
        ----------
        input: NDArray[(n,), np.uint]
            n-length vector of spike counts

        Keyword args
        ------------
        sample_time_ms: float
            Time measurement was taken in milliseconds

        Returns
        -------
        NDArray[(n,), float]
            n-length vector of firing rates
        """
        time_ms = kwargs["sample_time_ms"]
        if self.prev_rate is None:
            self.prev_rate = np.zeros(input.shape)
        if self.prev_time_ms is None:
            self.prev_time_ms = time_ms - self.T_s * 1000

        intersample_period_s = (time_ms - self.prev_time_ms) / 1000
        alpha = np.exp(-intersample_period_s / self.tau_s)
        curr_rate = self.prev_rate * alpha + (1 - alpha) * input / intersample_period_s
        self.prev_rate = curr_rate
        self.prev_time_ms = time_ms
        return curr_rate



# auto-import submodules
import cleosim.electrodes
import cleosim.opto
import cleosim.coordinates
import cleosim.stimulators
import cleosim.recorders
import cleosim.processing
import cleosim.utilities
import cleosim.visualization

from cleosim.base import (
    CLSimulator,
    Recorder,
    Stimulator,
    InterfaceDevice,
    IOProcessor,
)



"""Classes facilitating adding latency to :class:`~cleosim.processing.ProcessingBlock` computations"""
from abc import ABC, abstractmethod
import warnings

import numpy as np


class Delay(ABC):
    """Abstract base class for computing delays."""

    @abstractmethod
    def compute(self) -> float:
        """Compute delay."""
        pass


class ConstantDelay(Delay):
    """Simply adds a constant delay to the computation"""

    def __init__(self, delay_ms: float):
        """
        Parameters
        ----------
        delay_ms : float
            Desired delay in milliseconds
        """
        self.delay = delay_ms

    def compute(self):
        return self.delay


class GaussianDelay(Delay):
    """Generates normal-distributed delay.

    Will return 0 when a negative value is sampled.
    """

    def __init__(self, loc: float, scale: float):
        """
        Parameters
        ----------
        loc : float
            Center of distribution
        scale : float
            Standard deviation of delay distribution
        """
        self.loc = loc
        self.scale = scale

    def compute(self) -> float:
        out = np.random.normal(self.loc, self.scale)
        if out < 0:
            warnings.warn("Negative delay sampled. Returning 0.")
            out = 0
        return out



"""Contains basic stimulators."""
from brian2 import Unit
from cleosim.base import Stimulator


class StateVariableSetter(Stimulator):
    """Sets the given state variable of target neuron groups."""

    def __init__(
        self, name: str, variable_to_ctrl: str, unit: Unit, start_value: float = 0
    ):
        """
        Parameters
        ----------
        name : str
            Unique device name
        variable_to_ctrl : str
            Name of state variable to control
        unit : Unit
            Unit of that state variable: will be used in :meth:`update`
        start_value : float, optional
            Starting variable value, by default 0
        """
        super().__init__(name, start_value)
        self.neuron_groups = []
        self.var = variable_to_ctrl
        self.unit = unit
        self.value = start_value

    def connect_to_neuron_group(self, neuron_group):
        self.neuron_groups.append(neuron_group)
        self.update(self.value)

    def update(self, ctrl_signal: float) -> None:
        """Set state variable of target neuron groups

        Parameters
        ----------
        ctrl_signal : float
            Value to update variable to, without unit. The unit
            provided on initialization is automatically multiplied.
        """
        for ng in self.neuron_groups:
            setattr(ng, self.var, ctrl_signal * self.unit)



"""Basic processor and processing block definitions"""
from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Tuple, Any
from collections import deque

import numpy as np

from cleosim.base import IOProcessor
from cleosim.processing.delays import Delay


class ProcessingBlock(ABC):
    """Abstract signal processing stage or control block."""

    delay: Delay
    """The delay object determining compute latency for the block"""
    save_history: bool
    """Whether to record :attr:`t_in_ms`, :attr:`t_out_ms`, 
    and :attr:`values` with every timestep"""
    t_in_ms: list[float]
    """The walltime the block received each input.
    Only recorded if :attr:`save_history`"""
    t_out_ms: list[float]
    """The walltime of each of the block's outputs.
    Only recorded if :attr:`save_history`"""
    values: list[Any]
    """Each of the block's outputs.
    Only recorded if :attr:`save_history`"""

    def __init__(self, **kwargs):
        """
        It's important to use `super().__init__(**kwargs)` in the base class
        to use the parent-class logic here.

        Keyword args
        ------------
        delay : Delay
            Delay object which adds to the compute time

        Raises
        ------
        TypeError
            When `delay` is not a `Delay` object.
        """
        self.delay = kwargs.get("delay", None)
        if not isinstance(self.delay, Delay):
            raise TypeError("delay must be of the Delay class")
        self.save_history = kwargs.get("save_history", False)
        if self.save_history is True:
            self.t_in_ms = []
            self.t_out_ms = []
            self.values = []

    def process(self, input: Any, t_in_ms: float, **kwargs) -> Tuple[Any, float]:
        """Computes and saves output and output time given input and input time.

        The user should implement :meth:`~compute_output()` for their child
        classes, which performs the computation itself without regards for
        timing or saving variables.

        Parameters
        ----------
        input : Any
            Data to be processed
        t_in_ms : float
            Time the block receives the input data
        **kwargs : Any
            Key-value list of arguments passed to :func:`~compute_output()`

        Returns
        -------
        Tuple[Any, float]
            output, out time in milliseconds
        """
        out = self.compute_output(input, **kwargs)
        if self.delay is not None:
            t_out_ms = t_in_ms + self.delay.compute()
        else:
            t_out_ms = t_in_ms
        if self.save_history:
            self.t_in_ms.append(t_in_ms)
            self.t_out_ms.append(t_out_ms)
            self.values.append(out)
        return (out, t_out_ms)

    @abstractmethod
    def compute_output(self, input: Any, **kwargs) -> Any:
        """Computes output for given input.

        This is where the user will implement the desired functionality
        of the `ProcessingBlock` without regard for latency.

        Parameters
        ----------
        input : Any
            Data to be processed. Passed in from :meth:`process`.
        **kwargs : Any
            optional key-value argument pairs passed from
            :meth:`process`. Could be used to pass in such values as
            the IO processor's walltime or the measurement time for time-
            dependent functions.

        Returns
        -------
        Any
            output
        """
        pass


class LatencyIOProcessor(IOProcessor):
    """IOProcessor capable of delivering stimulation some time after measurement."""

    t_samp_ms: list[float]
    """Record of sampling times---each time :meth:`~put_state` is called."""

    def __init__(self, sample_period_ms: float, **kwargs):
        """
        Parameters
        ----------
        sample_period_ms : float
            Determines how frequently samples are taken from the network.

        Keyword args
        ------------
        sampling : str
            "fixed" or "when idle"; "fixed" by default

            "fixed" sampling means samples are taken on a fixed schedule,
            with no exceptions.

            "when idle" sampling means no samples are taken before the previous
            sample's output has been delivered. A sample is taken ASAP
            after an over-period computation: otherwise remains on schedule.

        processing : str
            "parallel" or "serial"; "parallel" by default

            "parallel" computes the output time by adding the delay for a sample
            onto the sample time, so if the delay is 2 ms, for example, while the
            sample period is only 1 ms, some of the processing is happening in
            parallel. Output order matches input order even if the computed
            output time for a sample is sooner than that for a previous
            sample.

            "serial" computes the output time by adding the delay for a sample
            onto the output time of the previous sample, rather than the sampling
            time. Note this may be of limited
            utility because it essentially means the *entire* round trip
            cannot be in parallel at all. More realistic is that simply
            each block or phase of computation must be serial. If anyone
            cares enough about this, it will have to be implemented in the
            future.

        Note
        ----
        It doesn't make much sense to combine parallel computation
        with "when idle" sampling, because "when idle" sampling only produces
        one sample at a time to process.

        Raises
        ------
        ValueError
            For invalid `sampling` or `processing` kwargs
        """
        self.t_samp_ms = []
        self.out_buffer = deque([])
        self.sample_period_ms = sample_period_ms
        self.sampling = kwargs.get("sampling", "fixed")
        if self.sampling not in ["fixed", "when idle"]:
            raise ValueError("Invalid sampling scheme:", self.sampling)
        self.processing = kwargs.get("processing", "parallel")
        if self.processing not in ["serial", "parallel"]:
            raise ValueError("Invalid processing scheme:", self.processing)

    def put_state(self, state_dict: dict, sample_time_ms):
        self.t_samp_ms.append(sample_time_ms)
        out, t_out_ms = self.process(state_dict, sample_time_ms)
        if self.processing == "serial" and len(self.out_buffer) > 0:
            prev_t_out_ms = self.out_buffer[-1][1]
            # add delay onto the output time of the last computation
            t_out_ms = prev_t_out_ms + t_out_ms - sample_time_ms
        self.out_buffer.append((out, t_out_ms))
        self._needs_off_schedule_sample = False

    def get_ctrl_signal(self, query_time_ms):
        if len(self.out_buffer) == 0:
            return None
        next_out_signal, next_t_out_ms = self.out_buffer[0]
        if query_time_ms >= next_t_out_ms:
            self.out_buffer.popleft()
            return next_out_signal
        else:
            return None

    def _is_currently_idle(self, query_time_ms):
        return len(self.out_buffer) == 0 or self.out_buffer[0][1] <= query_time_ms

    def is_sampling_now(self, query_time_ms):
        if self.sampling == "fixed":
            if np.isclose(query_time_ms % self.sample_period_ms, 0):
                return True
        elif self.sampling == "when idle":
            if query_time_ms % self.sample_period_ms == 0:
                if self._is_currently_idle(query_time_ms):
                    self._needs_off_schedule_sample = False
                    return True
                else:  # if not done computing
                    self._needs_off_schedule_sample = True
                    return False
            else:
                # off-schedule, only sample if the last sampling period
                # was missed (there was an overrun)
                return self._needs_off_schedule_sample and self._is_currently_idle(
                    query_time_ms
                )
        return False

    @abstractmethod
    def process(self, state_dict: dict, sample_time_ms: float) -> Tuple[dict, float]:
        """Process network state to generate output to update stimulators.

        This is the function the user must implement to define the signal processing
        pipeline.

        Parameters
        ----------
        state_dict : dict
            {`recorder_name`: `state`} dictionary from :func:`~cleosim.CLSimulator.get_state()`
        time_ms : float

        Returns
        -------
        Tuple[dict, float]
            {'stim_name': `ctrl_signal`} dictionary and output time in milliseconds.
        """
        pass


class RecordOnlyProcessor(LatencyIOProcessor):
    """Take samples without performing any control.

    Use this if all you are doing is recording."""

    def __init__(self, sample_period_ms, **kwargs):
        super().__init__(sample_period_ms, **kwargs)

    def process(self, state_dict: dict, sample_time_ms: float) -> Tuple[dict, float]:
        return ({}, sample_time_ms)
